/*
  Copyright (C) 2005 MARTIN Loc & VINCENT Cedric
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/** \mainpage Pilote de priphrique sous Linux

\section toc Table des matires

 - \ref page_intro
 - \ref page_amelioptim
  - \ref subsec_protec
  - \ref subsec_boucle
  - \ref subsec_schedule
  - \ref subsec_autres
  .
 - \ref page_valid
 .

*/

/** \page page_intro Introduction

Le sujet de ce projet est de crer un pilote de priphrique \em factice sous un noyau
Linux 2.6. Etant donn que l'on n'accdera pas directement  un \em vrai priphrique,
nous utiliserons la mmoire centrale afin de \em simuler son comportement. Ainsi, lorsque
l'on crira vers le priphrique, notre pilote devra enregistrer ces donnes, et
tre capable de nous les restituer lors d'une prochaine lecture. Il est  noter que l'criture se
fera dans un fichier spcial diffrent de celui de la lecture, et que notre pilote devra
tre capable de grer ainsi plusieurs \em paires de fichiers spciaux (une par priphrique).

Lors de son chargement, le pilote devra prendre en compte deux paramtres:
 - \b DeviceName, est le prfix du nom du priphrique apparaissant dans le rpertoire \em /dev
(car on utilise le systme DevFS)
 - \b DeviceCount, est le nombre de \em paires de priphriques que l'on utilisera. 
 .

Ainsi son chargement avec la commande :
\verbatim
insmod dummy.ko DeviceName="VinMart" DeviceCount=3
\endverbatim
produira les entres suivantes (en supposant que le systme nous ai assign le majeur 253) :
\verbatim
crw-rw-rw-  1 root root 253, 1 1970-01-01 01:00 /dev/VinMart_R0
crw-rw-rw-  1 root root 253, 3 1970-01-01 01:00 /dev/VinMart_R1
crw-rw-rw-  1 root root 253, 5 1970-01-01 01:00 /dev/VinMart_R2
crw-rw-rw-  1 root root 253, 0 1970-01-01 01:00 /dev/VinMart_W0
crw-rw-rw-  1 root root 253, 2 1970-01-01 01:00 /dev/VinMart_W1
crw-rw-rw-  1 root root 253, 4 1970-01-01 01:00 /dev/VinMart_W2
\endverbatim
Ainsi, ce qui est crit dans \em VinMart_Wx doit tre lisible dans \em VinMart_Rx mais pas
dans \em VinMart_Ry. Il faut prciser que la lecture est destructrive, c'est  dire que la
mmoire utilise pour stocker les donnes est libre lors de la lecture. Nous avons
pris la dcision que les fichiers spciaux avec un numro de mineur paire tait ceux d'criture, et donc
que les fichiers spciaux avec un numro de mineur impaire tait ceux de lecture.

En ce qui concerne ce rapport, nous ne nous attarderons pas sur <em>"comment faire un pilote de
priphrique sous Linux"</em>, puisqu'il s'agirait ni plus ni moins de paraphraser le cours.
En revanche, nous nous pencherons dans les sections suivantes sur ce que l'on considre
spcifique  notre travail personnel. Enfin, nous invitons le lecteur  consulter
la page de Doxygenation du source \ref dummy.c pour tout ce qui concerne
la description de notre code, ainsi que la page de Doxygenation de l'entte
\ref dummy.h pour tout ce qui concerne les structures de donnes et les
variables globales.

\warning Il faudra bien viter la confusion entre les priphriques et les fichiers spciaux.
Ce qui veut dire qu'il y a \em DeviceCount priphriques pour \em 2xDeviceCount fichiers spciaux.

\ref page_amelioptim "aller  la section suivante"
*/

/** \page page_amelioptim Amliorations et optimisations

Tout le long du dveloppement de ce pilote, nous avons remis en question le plus souvent
possible ce que nous tions en train de faire ( tel point que l'utilisation d'un serveur
CVS fut indispensable). Cette manire de travailler nous a permis d'apporter quelques 
amliorations par rapport au sujet demand et quelques optimisations par rapport  un code
\em simple. Nous allons donc discuter de ces \em personnalisations dans cette section.

\section subsec_protec Intgrit et oprations non-bloquantes

Afin d'assurer l'intgrit de la liste doublement chaine, o sont stockes les donnes de
l'utilisateur du priphrique, nous avons fais en sorte que lorsque le fichier spcial
de lecture (ou d'criture) est ouvert, il n'est pas possible qu'un autre processus puisse
l'ouvrir. Cela est ralis par le tableau d'entiers (utiliss comme boolen)
\ref already_opened qui est protg par le tableau de smaphores \ref open_critique. Leur taille
est gale au double du nombre de priphriques grs par le pilote (donc un lment par 
fichier spcial).

Bien videmment, il est tout  fait possible d'ouvrir le fichier spcial de lecture et celui
d'criture en mme temps. En revanche, pour les mme raisons que prcdemment,
 lorsqu'il y a une opration de lecture (respectivement d'criture) sur un priphrique,
il n'est pas possible de raliser une opration d'criture (respectivement de lecture).
Cela est ralis par la variable \em in_use de la structure \ref private_dummy
(propre  chaque priphrique), protg par le smaphore \em in_use_critique de cette mme
structure.

On pourrait tre amener  penser que cette politique est une limitation, mais en fait, en plus
de prserver l'intgrit de la liste doublement chaine o sont stockes les donnes de
l'utilisateur, cela prserve aussi l'intgrit des donnes elles-mme. En effet,
on peut facilement imaginer que si deux processus crivent en mme temps dans le priphrique
et que celui gre cette situation (en mlangeant le flux), les donnes ainsi stockes
n'auraient pas une grande signification...

Mais qu'en est-t-il vraiment lorsqu'une telle situation arrive, faut-il que le second
processus soit bloqu, ou non ? En fait, nous avons fais le choix de laisser le
choix  l'utilisateur :) ! Ainsi, en fonction du mode d'ouverture (paramtre \em 
flags de la fonction \em open(2)) le pilote ne se comportera pas de la mme manire
(bloquant ou non). Pour cela, nous vrifions simplement si l'utilisateur a positionn le
drapeau \em O_NONBLOCK.

\section subsec_boucle Listes et boucles

Un effort particulier a t fait pour l'optimisation des listes. Celles-ci sont doublement
chaines afin d'accder directement au dernier lment lors d'une criture (concatnation
des donnes). Le double chanage permet aussi d'assurer une gnricit, ce qui nous permet de
nous affranchir la gestion d'un pointeur ventuel sur le dernier lment \b et sur le premier.

Certes l'utilisation d'une liste doublement chaine n'est pas une mauvaise ide, mais ce n'est 
pas non plus \em remarquable... En fait, nous nous sommes concentr sur le traitement de ces
listes dans les boucles. En effet, nous avons fait en sorte de \em sortir un maximum de code
des boucles afin des les rendre plus rapide, car il est mauvais de rester trop longtemps
dans le code noyau sans rendre la main ! Nous avons donc fait en sorte que les tests de sortie
de boucle ne soit pas \em pollus par des cas particuliers. Nous pouvons citer pas exemple
le cas d'une criture, o le test pour savoir si la liste est vide est avant la boucle (et 
nous alloueons donc un lment en cas de besoin). Il faut bien noter qu'en plus d'enlever des
intructions de la boucle, cela enlve surtout des conditions qui peuvent tre 
dsastreuse en terme de performances sur certaines architectures...

\section subsec_schedule Rordonnacement du processus

Comme nous l'avons dit prcdemment, il est mauvais de rester dans le code noyau sans rendre
la main. En effet, supposons que l'on crive un fichier de 50 Mo en <b>une fois</b>,
notre pilote va boucler tant que \b toutes
les donnes ne seront pas stockes... Cette exemple prend 5 secondes sur un Pentium IV  2.6 GHz
(lorsque l'on met la taille des lments de la listes  1 octets). Pendant ces 5 secondes, il est
impossible de faire quoique ce soit, il n'est mme pas possible de faire bouger le pointeur de la
souris !!!

Certe, avec un noyau premptif cela pourrait tre rsolu, mais faut-il encore avoir
compil son noyau avec cette fonctionnalit... Nous n'avons fait aucune hypothse en ce qui
concerne cela, et nous avons choisi de grer cette situation par nous mme. C'est ainsi
que nous avons rajout des appels  l'ordonnanceur dans nos boucles. Les
rsultats sont  la hauteur de nos esprances, puisque pour le mme test, le systme est
compltement fonctionnel tout le long du stockage.

Il faut tout de mme noter que l'opration d'criture (ou de lecture) est alors plus lente,
mais cela ne veut pas dire que l'on reste plus longtemps dans le code noyau, bien au contraire.
Ce ralentissement du processus en cours d'opration est du au fait qu' on le rordonnance
souvent, mais le systme globale est infiniment plus rapide.

\section subsec_autres Mmoire dynamique

Nous allons parler rapidement dans cette section des petits dtails qui font que notre
pilote de priphrique fontionne (en thorie) plutt bien :)

Afin de supprimer les bogues lis  l'utilisation de la mmoire dynamique, nous avons fait
en sorte de limiter les appels  \em kmalloc et \em kfree. Ainsi, tout ce qui pouvait
tre fait en \em static l'a t, et lorsque l'on pouvait grouper ces appels, nous ne nous 
sommes pas priv. Par exemple, au dbut nous utilisions un tableau (dynamique) d'entier pour
savoir si le priprique tait occup, puis nous avons remplac ce systme en ajoutant
la variable \em in_use (qui a le mme rle) dans la structure \ref private_dummy .
Etant donn qu'il existe un tableau de cette structure, nomm \ref device_fifo, qui
a exactement la bonne taille
(puisqu'il a la mme correspondante entre ces lments et les priphriques), nous avons
donc un code plus \em propre (tout ce qui concerne un priphrique est dans une unique
structure) et surtout nous nous affranchissons d'une allocation et d'une libration (qui sont
souvent l'origine d'un bogue).

Toujours en ce qui concerne la mmoire dynamique, nous prenons soin de mettre les pointeurs
 \em NULL lorsqu'on les dsalloue. Ainsi, si on le rutilise par mgarde, on est prvenu
immdiatement :) !

Il y a encore d'autre dtails que l'on a oubli de mentionner (par exemple que l'on vrifie
\b tous les retours de fonctions) ou qui n'ont pas une grande
utilit par rapport  la comprhension de notre projet (par exemple le respect du 
Linux Kernel Coding Style), mais nous prfrons ne pas ennuyer le lecteur d'avantage :) !

\ref page_valid "aller  la section suivante"
*/

/** \page page_valid Validations

Un effort particulier a t fait sur la validation de notre pilote. Nous avons par exemple
utilis \em md5sum pour vrifier l'intgrit des donnes lues en comparant le rsultat avec
ce qui avait t crit dans le priphrique. De mme, nous avons utilis \em dd pour lire ou
crire un nombre \b prcis d'octet.

Mise  part l'utilisation d'outils existants, nous avons crit nos propres petits programmes
de tests surtout afin de tester toutes les amliorations que l'on avait faites par rapport au
sujet original. De mme, nous avons utilis nos propres petits scripts afin de vrifier le
comportement de notre pilote lors d'accs concurrenciels.

Mais ce qui nous a surtout permis de valider notre code, c'est une relecture \b systmatique
de ce que chacun faisait, pas une ligne est trangre  un lment de notre binme. Ainsi,
lorsque l'on explique son code  son camarade, on remarque souvent des erreurs d'tourderies
:) !

\author MARTIN Loc & VINCENT Cedric

Aller  la Doxygenation : \ref dummy.c, \ref dummy.h
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/slab.h>
#include "dummy.h"

/** Fonction d'ouverture des fichiers speciaux.
    Nous vrifions dans un premier temps si le mineur du fichier spcial est valide,
    c'est  dire qu'il appartient bien  ceux gr par notre pilote. Ensuite, nous
    prenons soin de vrifier que le fichier spcial n'est pas dj ouvert (par un autre
    processus par exemple), auquel cas nous indiquons que la ressource est occupe.
    Dans un troisime temps nous empchons qu'un fichier spcial de lecture ne soit
    ouvert en criture, et vice-versa. Puis nous dclarons au noyau que le priphrique
    est utilis (en incrmentant le compteur d'usage), et enfin nous associons le
    \em private_data du fichier spcial avec son \ref private_dummy correspondant.

    \return 0 en cas de succs, -errno sinon.
*/
static int dummy_open (struct inode *inode, struct file *file)
{
	int minor_lu = -1;
	int index = 0;

	// recupere le numero de mineur du fichier special
	minor_lu = MINOR(inode->i_rdev);

	// verifie qu'il est valide
	if (minor_lu > (DeviceCount*2 - 1)) {
		DEBUG && printk (KERN_ALERT "peripherique non supporte !\n");
		return -ENODEV;
	}

	// debut de section critique (prise de semaphore)
	if( down_interruptible(&open_critique[minor_lu]) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}

	// verifie que quelqu'un n'est pas en train d'utiliser le fichier special
	if(1 == already_opened[minor_lu]) {
		DEBUG && printk (KERN_ALERT "deja ouvert !\n");
		up(&open_critique[minor_lu]);
		return -EBUSY; //errno : Resource busy
	}	
	already_opened[minor_lu] = 1;

	// verifie que sur le fichier special ecriture on ne demande pas de lecture
	if (0 == minor_lu % 2 && file->f_mode & FMODE_READ) {
		DEBUG && printk (KERN_ALERT "ouverture non autorisee en lecture !\n");
		up(&open_critique[minor_lu]);
		return -EACCES;
	}

	// verifie que sur le fichier special de lecture on ne demande pas d'ecriture  
	if (1 == minor_lu % 2 && file->f_mode & FMODE_WRITE) {
		DEBUG && printk (KERN_ALERT "ouverture non autorisee en ecriture !\n");
		up(&open_critique[minor_lu]);
		return -EACCES;
	}

	// incremente le compteur d'usage du module (premiere chose a faire)
	//MOD_INC_USE_COUNT;
	if(try_module_get(THIS_MODULE) < 0) {
		DEBUG && printk (KERN_ALERT "erreur d'incrementation du compteur d'usage\n");
		up(&open_critique[minor_lu]);
		return -ERESTARTSYS;
	}

	// positionne le pointeur de pile sur le bon peripherique
	index = (int)minor_lu/2;
	file->private_data = &device_fifo[index];	

	// fin de section critique (relachement du semaphore)
	up(&open_critique[minor_lu]);

	DEBUG && printk (KERN_ALERT "ouverture autorisee !\n");
	DEBUG && printk (KERN_ALERT "utilisation de la file %d\n", index);
	
	return 0;
}

/** Fonction de fermeture des fichiers speciaux.
    On indique simplement que le fichier special est  nouveau accessible et nous dcrmentons
    le compteur d'usage.

    \return 0 en cas de succs, -errno sinon.
*/
static int dummy_release (struct inode *inode, struct file *file)
{
	int minor_lu = -1;
	int index = 0;

	// recupere le numero de mineur du fichier special
	minor_lu = MINOR(inode->i_rdev);
	index = (int)minor_lu/2;

	DEBUG && printk (KERN_ALERT "fichier special ferme !\n");	

	if( down_interruptible(&open_critique[minor_lu]) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}
	already_opened[minor_lu] = 0;

	module_put(THIS_MODULE); // (derniere chose a faire)
	//MOD_DEC_USE_COUNT;
	up(&open_critique[minor_lu]);

	return 0;
}

/** Fonction d'ecriture sur les peripheriques.
    Tout d'abord, nous regardons si le priphrique est dj occup, si c'est le cas
    et que l'utilisateur a demand un accs non bloquant, nous quittons la fonction en prcisant
    que la resource est momentanment inaccssible, sinon nous bloquons le processus.
    En ce qui concerne le remplissage de la liste des donnes, nous vrifions dans un premier
    temps que la liste n'est pas vide, si tel n'est pas le cas, nous alloueons un premier lment
    puis nous rentrons dans la boucle de remplissage. Dans cette dernire, nous rordonnanons
    le processus afin de librer le systme.

    \return 0 en cas de succs, -errno sinon.
*/
static ssize_t dummy_write (struct file *file, const char *buffer, size_t size, loff_t *offset)
{	
	struct private_dummy *private_data = file->private_data;
	struct element_fifo *last_element = NULL;
	struct element_fifo *new_element = NULL;
	struct element_fifo *current_element = NULL;
	struct element_fifo *head_element = private_data->head;
	size_t count = 0; // donnees ecrites par element
	size_t total_length = 0; // total des donnees ecrites
	unsigned long ret_code = 0;
	size_t origin_size = size; // total des donnees a ecrire
	size_t retvalue = 0; // code de retour

	if( down_interruptible(&private_data->in_use_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}

	// gestion de l'ecriture non bloquante
	if (1 == private_data->in_use && O_NONBLOCK & file->f_flags) {
		up(&private_data->in_use_critique);
		return -EAGAIN;
	} 
        private_data->in_use = 1;
	up(&private_data->in_use_critique);

	// debut de section critique (prise de semaphore)
	if( down_interruptible(&private_data->rw_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}

	// s'il n'y pas de tete , il faut la creer et l'initialiser (elle pointe (next et prev)
	// sur elle-meme)
	if(NULL == head_element) {
		head_element = kmalloc (sizeof(struct element_fifo), GFP_KERNEL);
		if(NULL == head_element) {
			DEBUG && printk (KERN_ALERT "allocation ratee");
			return -ERESTARTSYS;
		}

		DEBUG && printk(KERN_ALERT "allocation de l'element de tete\n");
	
		head_element->used_bytes = 0;
		head_element->position = 0;
		head_element->prev = head_element;
		head_element->next = head_element;

		private_data->head = head_element;
	}

	// tant que les donnees ne sont pas toutes ecrites
	while(total_length < origin_size) {
		// se positionner sur le dernier element.
		current_element = head_element->prev;		
	
		// verifie que le buffer a ecrire rentre dans la place restante du bloc
		if (size < BUF_SIZE - current_element->used_bytes) {
			count = size;
		} else {
			//sinon completer l'element avant d'en creer un autre
			count = BUF_SIZE - current_element->used_bytes;
		}

		// ecriture de 'count' caracteres du buffer dans l'element en cours
		ret_code = copy_from_user(current_element->buffer + current_element->used_bytes,
					  buffer, count);

		if (0 != ret_code) {
			DEBUG && printk(KERN_ALERT "erreur d'ecriture\n");
			retvalue = -EFAULT;
			goto ret_write;
		} 

		DEBUG && printk(KERN_ALERT "ecriture de %d octets sur %d\n", (int)count, 
				(int)size);

		// deplacement du pointeur sur le buffer du nombre de caracteres deja ecrits
		buffer = buffer + count;
		// mettre a jour le compteur de ce qu'il reste a ecrire
		size = size - count;
		// mettre a jour des compteurs de ce qui est ecrit
		total_length = total_length + count;
		current_element->used_bytes = count + current_element->used_bytes;

		DEBUG && printk(KERN_ALERT
				"occupation du noeud %d octets sur %d dans la file %d\n",
				current_element->used_bytes, BUF_SIZE, private_data->index);

 		if (BUF_SIZE == current_element->used_bytes) { 
 			// allocation d'un nouvel element 
 			new_element = kmalloc (sizeof(struct element_fifo), GFP_KERNEL); 
 			if (new_element == NULL)  { 
 				DEBUG && printk (KERN_ALERT "allocation ratee");
				retvalue = -ERESTARTSYS;
				goto ret_write;
 			} 

			// creation des liens entre le nouvel element et le reste de la pile
 			last_element = head_element->prev; 
 			last_element->next = new_element; 
 			head_element->prev = new_element; 

			// initialisation du nouvel element
 			new_element->used_bytes = 0; 
 			new_element->position = 0; 
 			new_element->prev = last_element; 
 			new_element->next = head_element; 

 			DEBUG && printk(KERN_ALERT "allocation d'un nouveau element\n"); 
 		} 
		
		// reordonnancement du processus, afin de liberer le systeme.
		schedule();
	}	
	retvalue = total_length;

ret_write:
	// fin de section critique (relachement du semaphore)
	up(&private_data->rw_critique);

	if( down_interruptible(&private_data->in_use_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}
        private_data->in_use = 0;
	up(&private_data->in_use_critique);

	return retvalue;
}

/** Fonction de lecture sur les peripheriques.
    Tout d'abord, nous regardons si le priphrique est dj occup, si c'est le cas
    et que l'utilisateur a demand un accs non bloquant, nous quittons la fonction en prcisant
    que la resource est momentanment inaccssible, sinon nous bloquons le processus.
    Dans la lecture, nous rordonnanons le processus afin de librer le systme.

    \return 0 en cas de succs, -errno sinon.
*/
static ssize_t dummy_read (struct file *file, char *buffer, size_t size, loff_t *offset)
{
	struct private_dummy *private_data = file->private_data;
	struct element_fifo *head_element = private_data->head;
	struct element_fifo *current_element = NULL;
	struct element_fifo *next_element = NULL;
	struct element_fifo *prev_element = NULL;
	size_t count = 0; // donnees lues par elements
	size_t total_length = 0; // total des donnees lues
	unsigned long ret_code = 0;
	size_t origin_size = size; // total des donnees a lire
	size_t retvalue = 0; // code de retour

	if( down_interruptible(&private_data->in_use_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}

	// gestion de la lecture non bloquante
	if (1 == private_data->in_use && O_NONBLOCK & file->f_flags) {
		up(&private_data->in_use_critique);
		return -EAGAIN;
	} 

        private_data->in_use = 1;

	up(&private_data->in_use_critique);

	// debut de section critique (prise de semaphore)       
 	if( down_interruptible(&private_data->rw_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}

        //tant que le message n'est pas lu en entier ou qu'il n'y a plus rien a lire 
 	while(total_length < origin_size && NULL != head_element) { 
 		current_element = head_element; 

		// verifie que le buffer a lire est plus petit qu'un bloc
 		if (size < current_element->used_bytes) { 
 			count = size; 
 		} else { // sinon lire la taille du bloc
 			count = current_element->used_bytes; 
 		} 

		// lecture de 'count' caracteres de l'element en cours a partir de 'position'
 		ret_code = copy_to_user(buffer,
					current_element->buffer+current_element->position,
					count); 

 		if (0 != ret_code) { 
 			DEBUG && printk(KERN_ALERT "erreur de lecture\n");
			retvalue = -EFAULT; 
 			goto ret_read;
 		}  

 		DEBUG && printk(KERN_ALERT "lecture de %d octets sur %d\n", (int)count,
				(int)size); 

		// mise a jour des pointeurs de buffer et du nombre de caracteres lus
 		buffer = buffer + count; 
 		size = size - count; 
 		total_length = total_length + count; 
 		current_element->used_bytes = current_element->used_bytes - count;
		current_element->position = current_element->position + count;
		DEBUG && printk(KERN_ALERT
				"occupation du noeud %d octets sur %d dans la file %d\n",
				current_element->used_bytes, BUF_SIZE, private_data->index);

 		//si element vide , desallouer la memoire sauf si c'est le dernier element  
 		if(0 == current_element->used_bytes) {
			prev_element = current_element->prev;
			next_element = current_element->next;

			prev_element->next = next_element;
			next_element->prev = prev_element;

			if(current_element != current_element->next) {
				head_element = next_element;
			} else {
				head_element = NULL;
			}
			
			private_data->head = head_element;
 
			kfree(current_element);			
			current_element = NULL;
		}
		
		// force le reordonnacement du processus
		schedule();
 	} 
	retvalue = total_length;

ret_read:
	// fin de section critique (relachement du semaphore)
 	up(&private_data->rw_critique);

	if( down_interruptible(&private_data->in_use_critique) ) {
		DEBUG && printk (KERN_ALERT "impossible de prendre le semaphore\n");
		return -ERESTARTSYS;
	}
        private_data->in_use = 0;
	up(&private_data->in_use_critique);

	return retvalue;
} 

/** Fonction de cration et d'initialisation des pripheriques.
    Dans un premier temps, nous enregistrons notre pilote auprs du noyau et nous lui 
    demandons de nous affilier un numro de majeur.
    Puis nous crons (grce au systme DevFS) les fichiers spciaux dans \em /dev.
    Enfin, nous alloueons et initialisons les diffrents tableaux ncessaires au pilote.

    \return 0 en cas de succs, -errno sinon.
*/
static int dummy_init(void)
{
	int result;
	int i;

	// allocation dynamique d'un majeur pour le pilote dummy
	result = register_chrdev(DUMMY_MAJOR, DeviceName, &dummy_fops);
	if (result < 0) {
		DEBUG && printk(KERN_ALERT "majeur %d indisponible\n", result);
		return -ERESTARTSYS;
	}
	// recuperation du numero de majeur pour la creation des peripheriques
	if (0 == dummy_major) {
		dummy_major = result;
	}

	// creation des peripheriques
	for (i = 0; i < DeviceCount; i++) {
	        // mineur pair => ecriture
		devfs_mk_cdev(MKDEV(dummy_major, 2*i), S_IFCHR|S_IRUGO|S_IWUGO, "%s_W%d",
			      DeviceName, i);
		DEBUG && printk(KERN_ALERT "%s_R%d cree\n", DeviceName, i);

	        // mineur impair => lecture
		devfs_mk_cdev(MKDEV(dummy_major, 2*i+1), S_IFCHR|S_IRUGO|S_IWUGO, "%s_R%d",
			      DeviceName, i);
		DEBUG && printk(KERN_ALERT "%s_W%d cree\n", DeviceName, i);
	}

	// allocation memoire du tableau d'etat des fichiers speciaux ouverts
	already_opened = kmalloc (DeviceCount*2*sizeof(int), GFP_KERNEL);
	if (already_opened == NULL)  {
		DEBUG && printk (KERN_ALERT "allocation ratee");
		return -ERESTARTSYS; 
	}

	// allocation memoire du semaphore de protection du tableau d'etat des fichier speciaux ouverts
	open_critique = kmalloc (DeviceCount*2*sizeof(struct semaphore), GFP_KERNEL);
	if (open_critique == NULL)  {
		DEBUG && printk (KERN_ALERT "allocation ratee");
		kfree(already_opened);
		return -ERESTARTSYS; 
	}

	// inutile d'utiliser le semaphore (de protection) pour cette operation car il ne peut pas y avoir d'acces concurrent
	// initialisation du tableau d'etat des fichier speciaux ouvrerts et des sempahores de protection
	for(i = 0; i < DeviceCount*2 ; i++) {
		already_opened[i] = 0;
	        sema_init(&open_critique[i], 1);
	}

	// allocation et initialisation des pointeurs sur la structure de pile (un par peripherique)
	device_fifo = kmalloc (DeviceCount*sizeof(struct private_dummy), GFP_KERNEL);
	if(NULL == device_fifo) {
		DEBUG && printk (KERN_ALERT "allocation ratee");
		kfree(already_opened);
		kfree(open_critique);
		return -ERESTARTSYS; 
	}

	for(i = 0; i < DeviceCount; i++) {
		device_fifo[i].index = i;
		device_fifo[i].head  = NULL;
		device_fifo[i].in_use = 0;
		sema_init(&device_fifo[i].in_use_critique, 1);
		sema_init(&device_fifo[i].rw_critique, 1);
	}

	DEBUG && printk(KERN_ALERT "module %s charge gerant 2x%d peripheriques\n", DeviceName,
			DeviceCount);

	// Une valeur de retour differente de 0 signifie un echec de init_module et donc du
	// chargement du module !
	return 0;
}

/** Fonction de fermeture des priphriques.
    Dans un premier temps, nous librons la mmoire utilise par les tableaux allous et
    initialiss dans \ref dummy_init. Puis nous supprimons tous les lments sotcks dans
    les listes. Enfin nous supprimons les fichiers spciaux de \em /dev et nous dsenregistrons
    notre pilote auprs du noyau.
*/
static void dummy_cleanup(void)
{
	int i = 0;	
	int result;
	struct element_fifo *tmp_element = NULL;
	struct element_fifo *current_element = NULL;

	// liberation de la memoire du semaphore de protection du tableau d'etat des fichier speciaux ouverts
	kfree(open_critique);
	open_critique = NULL;

	kfree(already_opened);
	already_opened = NULL;
	
	// liberation de la memoire des piles de donnees
	for (i = 0; i < DeviceCount; i++) {
	        // recuperation de la tete de la pile
		current_element = device_fifo[i].head;
		
		if(NULL != current_element) { //si la tete existe
		        // tant qu'il y a au moins une boite en plus de la tete
			while(current_element != current_element->next) { 
			        // positionnement sur l'element suivant
			        tmp_element = current_element->next; 

				// suppression de liens de l'element courant
				tmp_element->prev =  current_element->prev;
				current_element->prev->next = tmp_element;
				
				// suppression de l'element courant
				kfree(current_element); 
				current_element = NULL; 

				// l'element courant devient l'ancien element suivant
				current_element = tmp_element; 
				DEBUG && printk (KERN_ALERT "supression d'un noeud !\n"); 
			}
		
			// suppression du dernier element
			kfree(current_element); 
			current_element = NULL;

			DEBUG && printk (KERN_ALERT "supression d'un noeud !\n"); 
		}
	}

	// liberation de la memoire des pointeurs sur la structure de pile (un par peripherique)
	kfree(device_fifo);
	device_fifo = NULL;
	
	// suppression des peripheriques
	for (i = 0; i < DeviceCount; i++) {
		devfs_remove("%s_R%d", DeviceName, i);
		DEBUG && printk(KERN_ALERT "%s_R%d supprime\n", DeviceName, i);
	
		devfs_remove("%s_W%d", DeviceName, i);
		DEBUG && printk(KERN_ALERT "%s_W%d supprime\n", DeviceName, i);
	}

	// desenregistrement du pilote au niveau du noyau
 	result = unregister_chrdev(dummy_major, "dummy"); 
 	if (result < 0) {
 		DEBUG && printk (KERN_ALERT "desenregistrement %d rate\n", dummy_major); 
 		return; 
 	} 

	DEBUG && printk(KERN_ALERT "dummy decharge\n");
}

module_init(dummy_init);
module_exit(dummy_cleanup);

MODULE_PARM (DeviceCount, "i");
MODULE_PARM (DeviceName, "s");
